@@ -10,6 +10,7 @@
 #define XSIMD_MATH_COMPLEX_HPP
 
 #include "../types/xsimd_complex_base.hpp"
+#include "xsimd_basic_math.hpp"
 #include "xsimd_exponential.hpp"
 #include "xsimd_hyperbolic.hpp"
 #include "xsimd_logarithm.hpp"
@@ -19,26 +20,65 @@
 namespace xsimd
 {
     template <class X>
-    typename X::real_batch abs(const simd_complex_batch<X>& z);
+    real_batch_type_t<X> real(const simd_base<X>& z);
 
     template <class X>
-    typename X::real_batch arg(const simd_complex_batch<X>& z);
+    real_batch_type_t<X> imag(const simd_base<X>& z);
 
     template <class X>
-    X conj(const simd_complex_batch<X>& z);
+    real_batch_type_t<X> arg(const simd_base<X>& z);
 
     template <class X>
-    X sqrt(const simd_complex_batch<X>& z);
+    batch_type_t<X> conj(const simd_base<X>& z);
 
     template <class X>
-    typename simd_batch_traits<X>::real_value_type
-    norm(const simd_complex_batch<X>& rhs);
+    real_batch_type_t<X> norm(const simd_base<X>& rhs);
 
     template <class X>
-    X proj(const simd_complex_batch<X>& rhs);
+    batch_type_t<X> proj(const simd_base<X>& rhs);
 
     namespace detail
     {
+        /********
+         * sign *
+         ********/
+
+        template <class T, std::size_t N>
+        inline batch<T, N> sign_complex_impl(const batch<T, N>& z)
+        {
+            using b_type = batch<T, N>;
+            using r_type = typename b_type::real_batch;
+            auto rz = z.real();
+            auto iz = z.imag();
+            return select(rz != r_type(0.),
+                          b_type(sign(rz)),
+                          b_type(sign(iz)));
+        }
+
+        template <class T, std::size_t N>
+        struct sign_impl<batch<std::complex<T>, N>, false>
+        {
+            using batch_type = batch<std::complex<T>, N>;
+
+            static inline batch_type compute(const batch_type& z)
+            {
+                return sign_complex_impl(z);
+            }
+        };
+
+#ifdef XSIMD_ENABLE_XTL_COMPLEX
+        template <class T, bool i3ec, std::size_t N>
+        struct sign_impl<batch<xtl::xcomplex<T, T, i3ec>, N>, false>
+        {
+            using batch_type = batch<xtl::xcomplex<T, T, i3ec>, N>;
+
+            static inline batch_type compute(const batch_type& z)
+            {
+                return sign_complex_impl(z);
+            }
+        };
+#endif
+
         /*******
          * exp *
          *******/
@@ -77,6 +117,46 @@ namespace xsimd
         };
 #endif
 
+        /*********
+         * expm1 *
+         *********/
+
+        template <class T, std::size_t N>
+        inline batch<T, N> expm1_complex_impl(const batch<T, N>& z)
+        {
+            using b_type = batch<T, N>;
+            using r_type = typename b_type::real_batch;
+            r_type isin = sin(z.imag());
+            r_type rem1 = expm1(z.real());
+            r_type re = rem1 + r_type(1.);
+            r_type si = sin(z.imag() * r_type(0.5));
+            return batch<T, N>(rem1 - r_type(2.) * re * si * si, re * isin);
+        }
+
+        template <class T, std::size_t N>
+        struct expm1_kernel<batch<std::complex<T>, N>, std::complex<T>>
+        {
+            using batch_type = batch<std::complex<T>, N>;
+
+            static inline batch_type compute(const batch_type& z)
+            {
+                return expm1_complex_impl(z);
+            }
+        };
+
+#ifdef XSIMD_ENABLE_XTL_COMPLEX
+        template <class T, bool i3ec, std::size_t N>
+        struct expm1_kernel<batch<xtl::xcomplex<T, T, i3ec>, N>, xtl::xcomplex<T, T, i3ec>>
+        {
+            using batch_type = batch<xtl::xcomplex<T, T, i3ec>, N>;
+
+            static inline batch_type compute(const batch_type& z)
+            {
+                return expm1_complex_impl(z);
+            }
+        };
+#endif
+
         /*******
          * log *
          *******/
@@ -111,17 +191,50 @@ namespace xsimd
         };
 #endif
 
+        /*********************
+         * logN_complex_impl *
+         *********************/
+
+        template <class T, std::size_t N>
+        inline batch<T, N> logN_complex_impl(const batch<T, N>& z, typename batch<T, N>::real_value_type base)
+        {
+            using b_type = batch<T, N>;
+            using rv_type = typename b_type::real_value_type;
+            return log(z) / b_type(rv_type(base));
+        }
+
+        /********
+         * log2 *
+         ********/
+
+        template <class T, std::size_t N>
+        struct log2_kernel<batch<std::complex<T>, N>, std::complex<T>>
+        {
+            using batch_type = batch<std::complex<T>, N>;
+
+            static inline batch_type compute(const batch_type& z)
+            {
+                return logN_complex_impl(z, std::log(2));
+            }
+        };
+
+#ifdef XSIMD_ENABLE_XTL_COMPLEX
+        template <class T, bool i3ec, std::size_t N>
+        struct log2_kernel<batch<xtl::xcomplex<T, T, i3ec>, N>, xtl::xcomplex<T, T, i3ec>>
+        {
+            using batch_type = batch<xtl::xcomplex<T, T, i3ec>, N>;
+
+            static inline batch_type compute(const batch_type& z)
+            {
+                return logN_complex_impl(z, std::log(2));
+            }
+        };
+#endif
+
         /*********
          * log10 *
          *********/
 
-        template <class T, std::size_t N>
-        inline batch<T, N> log10_complex_impl(const batch<T, N>& z)
-        {
-            using real_value_type = typename batch<T, N>::real_value_type;
-            return log(z) / batch<T, N>(real_value_type(std::log(10)));
-        }
-
         template <class T, std::size_t N>
         struct log10_kernel<batch<std::complex<T>, N>, std::complex<T>>
         {
@@ -129,7 +242,7 @@ namespace xsimd
 
             static inline batch_type compute(const batch_type& z)
             {
-                return log10_complex_impl(z);
+                return logN_complex_impl(z, std::log(10));
             }
         };
 
@@ -141,7 +254,49 @@ namespace xsimd
 
             static inline batch_type compute(const batch_type& z)
             {
-                return log10_complex_impl(z);
+                return logN_complex_impl(z, std::log(10));
+            }
+        };
+#endif
+
+        /*********
+         * log1p *
+         *********/
+
+        template <class T, std::size_t N>
+        inline batch<T, N> log1p_complex_impl(const batch<T, N>& z)
+        {
+            using b_type = batch<T, N>;
+            using r_type = typename b_type::real_batch;
+            b_type u = b_type(1.) + z;
+            b_type logu = log(u);
+            return select(u == b_type(1.),
+                          z,
+                          select(u.real() <= r_type(0.),
+                                 logu,
+                                 logu * z / (u - b_type(1.))));
+        }
+
+        template <class T, std::size_t N>
+        struct log1p_kernel<batch<std::complex<T>, N>, std::complex<T>>
+        {
+            using batch_type = batch<std::complex<T>, N>;
+
+            static inline batch_type compute(const batch_type& z)
+            {
+                return log1p_complex_impl(z);
+            }
+        };
+
+#ifdef XSIMD_ENABLE_XTL_COMPLEX
+        template <class T, bool i3ec, std::size_t N>
+        struct log1p_kernel<batch<xtl::xcomplex<T, T, i3ec>, N>, xtl::xcomplex<T, T, i3ec>>
+        {
+            using batch_type = batch<xtl::xcomplex<T, T, i3ec>, N>;
+
+            static inline batch_type compute(const batch_type& z)
+            {
+                return log1p_complex_impl(z);
             }
         };
 #endif
@@ -230,7 +385,7 @@ namespace xsimd
             b_type winf(infinity<r_type>(), infinity<r_type>());
             r_type wreal = sin(2 * z.real()) / d;
             r_type wimag = sinh(2 * z.imag());
-            b_type wres = select(isinf(wimag), b_type(wreal, r_type(1.)), b_type(wreal, wimag / d));
+            b_type wres = select(xsimd::isinf(wimag), b_type(wreal, r_type(1.)), b_type(wreal, wimag / d));
             return select(d == r_type(0.), winf, wres);
         }
 
@@ -337,7 +492,9 @@ namespace xsimd
             num = x2 + num * num;
             r_type den = y - one;
             den = x2 + den * den;
-            b_type res(w, 0.25 * log(num / den));
+            b_type res = select((x == r_type(0.)) && (y == r_type(1.)),
+                                b_type(r_type(0.), infinity<r_type>()),
+                                b_type(w, 0.25 * log(num / den)));
             return res;
         }
 
@@ -609,24 +766,6 @@ namespace xsimd
 #endif
     }
 
-    template <class X>
-    inline typename X::real_batch abs(const simd_complex_batch<X>& z)
-    {
-        return hypot(z.real(), z.imag());
-    }
-
-    template <class X>
-    inline typename X::real_batch arg(const simd_complex_batch<X>& z)
-    {
-        return atan2(z.imag(), z.real());
-    }
-
-    template <class X>
-    inline X conj(const simd_complex_batch<X>& z)
-    {
-        return X(z.real(), -z.imag());
-    }
-
     namespace detail
     {
         template <class T>
@@ -660,52 +799,142 @@ namespace xsimd
         }
     }
 
-    template <class X>
-    inline X sqrt(const simd_complex_batch<X>& z)
+    namespace detail
     {
-        using real_batch = typename X::real_batch;
-        using rvt = typename real_batch::value_type;
-        real_batch x = z.real();
-        real_batch y = z.imag();
-        real_batch sqrt_x = sqrt(fabs(x));
-        real_batch sqrt_hy = sqrt(0.5 * fabs(y));
-        auto cond = (fabs(x) > real_batch(4.) || fabs(y) > real_batch(4.));
-        x = select(cond, x * 0.25, x * detail::csqrt_scale_factor<rvt>());
-        y = select(cond, y * 0.25, y * detail::csqrt_scale_factor<rvt>());
-        real_batch scale = select(cond, real_batch(2.), real_batch(detail::csqrt_scale<rvt>()));
-        real_batch r = abs(X(x, y));
+        template <class T, std::size_t N>
+        struct complex_batch_kernel
+        {
+            using batch_type = batch<T, N>;
+            using real_batch = typename batch_type::real_batch;
 
-        auto condxp = x > real_batch(0.);
-        real_batch t0 = select(condxp, sqrt(0.5 * (r + x)), sqrt(0.5 * (r - x)));
-        real_batch r0 = scale * fabs((0.5 * y) / t0);
-        t0 *= scale;
-        real_batch t = select(condxp, t0, r0);
-        r = select(condxp, r0, t0);
-        X resg = select(y < real_batch(0.), X(t, -r), X(t, r));
-        real_batch ze(0.);
+            static real_batch abs(const batch_type& z)
+            {
+                return hypot(z.real(), z.imag());
+            }
 
-        return select(y == ze,
-                      select(x == ze,
-                             X(ze, ze),
-                             select(x < ze, X(ze, sqrt_x), X(sqrt_x, ze))),
-                      select(x == ze,
-                             select(y > ze, X(sqrt_hy, sqrt_hy), X(sqrt_hy, -sqrt_hy)),
-                             resg));
+            static batch_type sqrt(const batch_type& z)
+            {
+                using rvt = typename real_batch::value_type;
+                real_batch x = z.real();
+                real_batch y = z.imag();
+                real_batch sqrt_x = xsimd::sqrt(fabs(x));
+                real_batch sqrt_hy = xsimd::sqrt(0.5 * fabs(y));
+                auto cond = (fabs(x) > real_batch(4.) || fabs(y) > real_batch(4.));
+                x = select(cond, x * 0.25, x * detail::csqrt_scale_factor<rvt>());
+                y = select(cond, y * 0.25, y * detail::csqrt_scale_factor<rvt>());
+                real_batch scale = select(cond, real_batch(2.), real_batch(detail::csqrt_scale<rvt>()));
+                real_batch r = abs(batch_type(x, y));
+
+                auto condxp = x > real_batch(0.);
+                real_batch t0 = select(condxp, xsimd::sqrt(0.5 * (r + x)), xsimd::sqrt(0.5 * (r - x)));
+                real_batch r0 = scale * fabs((0.5 * y) / t0);
+                t0 *= scale;
+                real_batch t = select(condxp, t0, r0);
+                r = select(condxp, r0, t0);
+                batch_type resg = select(y < real_batch(0.), batch_type(t, -r), batch_type(t, r));
+                real_batch ze(0.);
+
+                return select(y == ze,
+                              select(x == ze,
+                                     batch_type(ze, ze),
+                                     select(x < ze, batch_type(ze, sqrt_x), batch_type(sqrt_x, ze))),
+                              select(x == ze,
+                                     select(y > ze, batch_type(sqrt_hy, sqrt_hy), batch_type(sqrt_hy, -sqrt_hy)),
+                                     resg));
+            }
+        };
+
+        template <class T, std::size_t N>
+        struct batch_kernel<std::complex<T>, N>
+             : public complex_batch_kernel<std::complex<T>, N>
+        {
+        };
+
+        #ifdef XSIMD_ENABLE_XTL_COMPLEX
+        template <class T, bool i3ec, std::size_t N>
+        struct batch_kernel<xtl::xcomplex<T, T, i3ec>, N>
+             : public complex_batch_kernel<xtl::xcomplex<T, T, i3ec>, N>
+        {
+        };
+        #endif
+    }
+
+    namespace detail
+    {
+        template <class B, bool is_complex>
+        struct real_imag_kernel
+        {
+            using return_type = typename B::real_batch;
+
+            static return_type real(const B& z)
+            {
+                return z.real();
+            }
+
+            static return_type imag(const B& z)
+            {
+                return z.imag();
+            }
+        };
+
+        template <class B>
+        struct real_imag_kernel<B, false>
+        {
+            using return_type = B;
+
+            static return_type real(const B& z)
+            {
+                return z;
+            }
+
+            static return_type imag(const B& z)
+            {
+                return B(typename B::value_type(0));
+            }
+        };
     }
 
     template <class X>
-    inline typename simd_batch_traits<X>::real_batch
-    norm(const simd_complex_batch<X>& rhs)
+    inline real_batch_type_t<X> real(const simd_base<X>& z)
     {
-        return rhs.real() * rhs.real() + rhs.imag() * rhs.imag();
+        using batch_type = batch_type_t<X>;
+        constexpr bool is_cplx = detail::is_complex<typename batch_type::value_type>::value;
+        return detail::real_imag_kernel<batch_type, is_cplx>::real(z());
     }
 
     template <class X>
-    inline X proj(const simd_complex_batch<X>& rhs)
+    inline real_batch_type_t<X> imag(const simd_base<X>& z)
     {
+        using batch_type = batch_type_t<X>;
+        constexpr bool is_cplx = detail::is_complex<typename batch_type::value_type>::value;
+        return detail::real_imag_kernel<batch_type, is_cplx>::imag(z());
+    }
+
+    template <class X>
+    inline real_batch_type_t<X> arg(const simd_base<X>& z)
+    {
+        return atan2(imag(z), real(z));
+    }
+
+    template <class X>
+    inline batch_type_t<X> conj(const simd_base<X>& z)
+    {
+        return X(z().real(), -z().imag());
+    }
+
+    template <class X>
+    inline real_batch_type_t<X> norm(const simd_base<X>& rhs)
+    {
+        return real(rhs) * real(rhs) + imag(rhs) * imag(rhs);
+    }
+
+    template <class X>
+    inline batch_type_t<X> proj(const simd_base<X>& rhs)
+    {
+        using batch_type = batch_type_t<X>;
         using real_batch = typename simd_batch_traits<X>::real_batch;
-        auto cond = isinf(rhs.real()) || isinf(rhs.imag());
-        return select(cond, X(infinity<real_batch>(), copysign(real_batch(0.), rhs.imag())), rhs);
+        auto cond = xsimd::isinf(rhs().real()) || xsimd::isinf(rhs().imag());
+        return select(cond, batch_type(infinity<real_batch>(), copysign(real_batch(0.), rhs().imag())), rhs());
     }
 }
 
